#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import re
import warnings
import pandas as pd
from tqdm import tqdm
from Bio.SeqIO.FastaIO import SimpleFastaParser

from ..data import check_input
from ..data import check_input_multi
from ..transform.reannotate_positions import __generate_all_sequences
from ..transform.util import assert_data_type_same

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Tuple
from typing import List
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

CHAINS = list("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789")


def __get_proteins_and_positions(
    peptide: str, protein_db: Dict[str, Dict[str, str]], error_on_no_match: bool = False
) -> Tuple[List[str], List[int]]:
    r"""Retrieve matching protein chains and peptide positions for a specific peptide.

    Matches the specified peptide against the given protein database and returns all protein chains
    that contain the peptide, as well as the corresponding peptide positions in those protein chains.
    Uses 0-based indexing.

    Parameters
    ----------
    peptide : str
        Unmodified peptide sequence.
    protein_db : dict of dict of str, str
        A dictionary that maps protein chain ids to their fasta entries, which are dictionaries
        that map key "header" to the sequence header and key "sequence" to the sequence.
    error_on_no_match : bool, default = False
        Wether an error should be raised if the peptide matches to none of the proteins.

    Returns
    -------
    tuple of list of str, list of int
        List of protein chain ids, and list of peptide positions.

    Raises
    ------
    RuntimeError
        If the peptide could not be matched to any protein. Only raised if 'error_on_no_match'
        is set to True.

    Notes
    -----
    This function should not be called directly, it is called from ``to_alphalink2()``.

    Warnings
    --------
    Contrary to most functions in pyXLMS, this function uses 0-based indexing.
    """
    proteins = list()
    positions = list()
    for chain, item in protein_db.items():
        base_seq = item["sequence"]
        seqs = __generate_all_sequences(base_seq)
        for seq in seqs:
            if peptide in seq:
                for match in re.finditer(peptide, seq):
                    proteins.append(chain)
                    positions.append(match.start())
    if error_on_no_match and len(proteins) == 0:
        raise RuntimeError(f"No match found for peptide {peptide}!")
    return (proteins, positions)


def __protein_supported_by_crosslink(
    sequence: str, crosslinks: List[Dict[str, Any]]
) -> bool:
    r"""Check if a specific protein is supported by any of the given crosslinks.

    Checks if a specific protein that is given by its sequence is supported by any of the input
    crosslinks.

    Parameters
    ----------
    sequence : str
        The sequence of the protein.
    crosslinks : list of dict of str, any
        A list of crosslinks.

    Returns
    -------
    bool
        Returns True if the protein is supported by any crosslink, otherwise False.

    Notes
    -----
    This function should not be called directly, it is called from ``to_alphalink2()``.
    """
    seqs = __generate_all_sequences(sequence)
    for seq in seqs:
        for crosslink in crosslinks:
            if crosslink["alpha_peptide"] in seq:
                return True
            if crosslink["beta_peptide"] in seq:
                return True
    return False


def to_alphalink2(
    crosslinks: List[Dict[str, Any]],
    fasta: str | BinaryIO,
    annotated_fdr: float | List[float] = 0.01,
    try_use_annotated_fdr: bool = True,
    filename_prefix: Optional[str] = None,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Exports a list of crosslinks to AlphaLink2 format.

    Exports a list of crosslinks to AlphaLink2 format. The tool AlphaLink2 is accessible
    via the link
    `github.com/Rappsilber-Laboratory/AlphaLink2 <https://github.com/Rappsilber-Laboratory/AlphaLink2>`_.

    Parameters
    ----------
    crosslinks : list of dict of str, any
        A list of crosslinks to export.
    fasta : str, or file stream
        The name/path of the fasta file containing protein sequences or a file-like object/stream.
        Please keep in mind that AlphaLink2 supports a maximum of 62 proteins/chains!
    annotated_fdr : float, or list of float, default = 0.01
        Value(s) to use for the "FDR" column in the AlphaLink2 crosslink table. If a single float value
        is given, all crosslinks will be annotated with that constant FDR value. If a list of floats is
        given, it should be of equal length as the crosslinks and correspondingly contain FDR values for
        all crosslinks in the same order, e.g. the FDR value for the first crosslink should be the first
        value in the list as well, the second for the second crosslink, and so forth.
    try_use_annotated_fdr : bool, default = True
        If pyXLMS annotated FDR - e.g. via ``transform.annotate_fdr()`` - should be used for the "FDR"
        column in the AlphaLink2 crosslink table, if available. This will override the values given via
        parameter ``annotated_fdr`` and should be ``False`` if you are passing a custom list of FDR values!
    filename_prefix : str, or None, default = None
        If not None, the exported data will be written to files with the specified filename prefix.
        The full list of written files can be accessed via the returned dictionary.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict of str, any
        Returns a dictionary with key ``AlphaLink2 crosslinks`` containing the formatted crosslink input for AlphaLink2,
        with key ``AlphaLink2 FASTA`` containing the FASTA file content for AlphaLink2,
        with key ``AlphaLink2 DataFrame`` containing the exported crosslinks as a pandas DataFrame,
        and with key ``Exported files`` containing a list of filenames of all files that were
        written to disk.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    ValueError
        If the provided crosslinks contain no elements.
    ValueError
        If the length of annotated_fdr does not match the length of the input crosslinks.
        Only applies if annotated_fdr is given as a list.
    IndexError
        If the provided crosslinks match to more than 62 proteins/chains in the FASTA file.
    RuntimeError
        If one or more of the crosslinks could not be matched to the proteins/chains in the
        FASTA file. Only is raised if 'verbose' is set to 2.

    Notes
    -----
    Please note that the legacy PDB format and therefore also AlphaLink2 only supports a maximum
    of 62 proteins/chains. If crosslinks of more than 62 proteins/chains are given, an error will
    be thrown!

    Examples
    --------
    >>> from pyXLMS.pipelines import pipeline
    >>> from pyXLMS.transform import filter_proteins
    >>> from pyXLMS.exporter import to_alphalink2
    >>> pr = pipeline(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.pdResult",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> cas9 = filter_proteins(pr["crosslinks"], proteins=["Cas9"])["Both"]
    >>> _export = to_alphalink2(
    ...     cas9, fasta="data/_fasta/Cas9_plus10.fasta", filename_prefix="Cas9"
    ... )
    """
    _ok = check_input(crosslinks, "crosslinks", list, dict)
    _ok = check_input_multi(annotated_fdr, "annotated_fdr", [float, list], float)
    _ok = check_input(try_use_annotated_fdr, "try_use_annotated_fdr", bool)
    _ok = (
        check_input(filename_prefix, "filename_prefix", str)
        if filename_prefix is not None
        else True
    )
    _ok = check_input(verbose, "verbose", int)
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")
    if isinstance(annotated_fdr, list) and len(annotated_fdr) != len(crosslinks):
        raise ValueError(
            "Length of annotated_fdr does not match length of crosslinks! "
            + "When providing a list it needs to contain FDR values for every crosslink and therefore be of equal length!"
        )
    if len(crosslinks) == 0:
        raise ValueError("Provided crosslinks contain no elements!")
    if "data_type" not in crosslinks[0] or crosslinks[0]["data_type"] != "crosslink":
        raise TypeError(
            "Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!"
        )
    _ok = assert_data_type_same(crosslinks)
    protein_db = dict()
    # read fasta file
    fasta_items = list()
    if isinstance(fasta, str):
        with open(fasta, "r", encoding="utf-8") as f:
            for item in SimpleFastaParser(f):
                fasta_items.append(item)
    else:
        for item in SimpleFastaParser(fasta):
            fasta_items.append(item)
    if len(fasta_items) > len(CHAINS):
        raise IndexError(
            "Found more than the supported 62 proteins/chains in the fasta file! Please trim fasta file to a maximum of 62 sequences!"
        )
    id = 0
    for item in fasta_items:
        if __protein_supported_by_crosslink(item[1], crosslinks):
            protein_db[CHAINS[id]] = {"header": item[0], "sequence": item[1]}
            id += 1
    # prepare fdr values
    fdr_values = (
        annotated_fdr
        if isinstance(annotated_fdr, list)
        else [annotated_fdr for xl in crosslinks]
    )
    # output
    alphalink2_txt = ""
    alphalink2_df_dict = {
        "residueFrom": [],
        "chain1": [],
        "residueTo": [],
        "chain2": [],
        "FDR": [],
    }
    # export crosslinks
    for id, xl in tqdm(
        enumerate(crosslinks),
        total=len(crosslinks),
        desc="Exporting crosslinks to AlphaLink2...",
    ):
        proteins_a, pep_position0_proteins_a = __get_proteins_and_positions(
            xl["alpha_peptide"], protein_db, error_on_no_match=False
        )
        proteins_b, pep_position0_proteins_b = __get_proteins_and_positions(
            xl["beta_peptide"], protein_db, error_on_no_match=False
        )
        # skip if crosslink matches to none of the proteins in the fasta
        if len(proteins_a) == 0 or len(proteins_b) == 0:
            if verbose == 1:
                warnings.warn(
                    RuntimeWarning(
                        (
                            f"Could not find matching proteins in FASTA file for crosslink id {id}:{xl['alpha_peptide']}-{xl['beta_peptide']}! "
                            "This warning can be ignored if this is to be expected."
                        )
                    )
                )
            if verbose == 2:
                raise RuntimeError(
                    (
                        f"Could not find matching proteins in FASTA file for crosslink id {id}:{xl['alpha_peptide']}-{xl['beta_peptide']}! "
                        "If this is to be expected please set verbose level to either 1 or 0!"
                    )
                )
            continue
        for i in range(len(proteins_a)):
            for j in range(len(proteins_b)):
                residueFrom = (
                    pep_position0_proteins_a[i] + xl["alpha_peptide_crosslink_position"]
                )
                chain1 = proteins_a[i]
                residueTo = (
                    pep_position0_proteins_b[j] + xl["beta_peptide_crosslink_position"]
                )
                chain2 = proteins_b[j]
                FDR = fdr_values[id]
                if try_use_annotated_fdr:
                    if xl["additional_information"] is not None:
                        if "pyXLMS_annotated_FDR" in xl["additional_information"]:
                            if not pd.isna(
                                xl["additional_information"]["pyXLMS_annotated_FDR"]
                            ):
                                FDR = xl["additional_information"][
                                    "pyXLMS_annotated_FDR"
                                ]
                alphalink2_df_dict["residueFrom"].append(residueFrom)
                alphalink2_df_dict["chain1"].append(chain1)
                alphalink2_df_dict["residueTo"].append(residueTo)
                alphalink2_df_dict["chain2"].append(chain2)
                alphalink2_df_dict["FDR"].append(FDR)
                alphalink2_txt += f"{residueFrom} {chain1} {residueTo} {chain2} {FDR}\n"
    # create fasta
    alphalink2_fasta = ""
    for chain in CHAINS:
        if chain in protein_db:
            item = protein_db[chain]
            alphalink2_fasta += f">{chain}|{item['header']}\n{item['sequence']}\n"
    # create pandas dataframe
    alphalink2_df = pd.DataFrame(alphalink2_df_dict)
    # export files
    exported_files = list()
    if filename_prefix is not None:
        with open(filename_prefix + "_AlphaLink2.txt", "w", encoding="utf-8") as f:
            f.write(alphalink2_txt)
            f.close()
        exported_files.append(filename_prefix + "_AlphaLink2.txt")
        with open(filename_prefix + "_AlphaLink2.fasta", "w", encoding="utf-8") as f:
            f.write(alphalink2_fasta)
            f.close()
        exported_files.append(filename_prefix + "_AlphaLink2.fasta")
    # return exported files
    return {
        "AlphaLink2 crosslinks": alphalink2_txt,
        "AlphaLink2 FASTA": alphalink2_fasta,
        "AlphaLink2 DataFrame": alphalink2_df,
        "Exported files": exported_files,
    }



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


F1 = "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.mzid"
F2 = "data/xlinkx/XLpeplib_Beveridge_Lumos_DSSO_MS3.mzid"


@pytest.mark.slow
def test1():
    from pyXLMS import parser as p

    with open(F1, "rb") as f:
        pr = p.read_mzid(f, verbose=0)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "mzIdentML"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 786


@pytest.mark.slow
def test2():
    from pyXLMS import parser as p

    with open(F2, "rb") as f:
        pr = p.read_mzid(f, verbose=0)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "mzIdentML"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 823



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    with open("data/pyxlms/csm.txt", "r", encoding="utf-8") as f:
        parser_result = p.read_custom(f, format="csv")
        f.close()

    assert parser_result["data_type"] == "parser_result"
    assert parser_result["completeness"] == "partial"
    assert parser_result["search_engine"] == "Custom"
    assert parser_result["crosslink-spectrum-matches"] is not None
    assert parser_result["crosslinks"] is None

    csms = parser_result["crosslink-spectrum-matches"]
    assert len(csms) == 2

    first_csm = csms[0]
    last_csm = csms[-1]

    assert first_csm["data_type"] == "crosslink-spectrum-match"
    assert first_csm["completeness"] == "full"
    assert first_csm["alpha_peptide"] == "KPEPTIDE"
    assert mts(first_csm["alpha_modifications"]) == "(1:[DSS|138.06808])"
    assert first_csm["alpha_peptide_crosslink_position"] == 1
    assert first_csm["alpha_proteins"] == ["Cas9"]
    assert first_csm["alpha_proteins_crosslink_positions"] == [13]
    assert first_csm["alpha_proteins_peptide_positions"] == [13]
    assert first_csm["alpha_score"] == pytest.approx(87.53)
    assert not first_csm["alpha_decoy"]
    assert first_csm["beta_peptide"] == "PEPKTIDE"
    assert mts(first_csm["beta_modifications"]) == "(4:[DSS|138.06808])"
    assert first_csm["beta_peptide_crosslink_position"] == 4
    assert first_csm["beta_proteins"] == ["Cas9"]
    assert first_csm["beta_proteins_crosslink_positions"] == [17]
    assert first_csm["beta_proteins_peptide_positions"] == [14]
    assert first_csm["beta_score"] == pytest.approx(100.3)
    assert not first_csm["beta_decoy"]
    assert first_csm["crosslink_type"] == "intra"
    assert first_csm["score"] == pytest.approx(87.53)
    assert first_csm["spectrum_file"] == "S1.raw"
    assert first_csm["scan_nr"] == 1
    assert first_csm["charge"] == 3
    assert first_csm["retention_time"] == pytest.approx(14.3)
    assert first_csm["ion_mobility"] == pytest.approx(50.0)

    assert last_csm["data_type"] == "crosslink-spectrum-match"
    assert last_csm["completeness"] == "full"
    assert last_csm["alpha_peptide"] == "EKTIDEM"
    assert (
        mts(last_csm["alpha_modifications"])
        == "(2:[DSS|138.06808]);(7:[Oxidation|15.994915])"
    )
    assert last_csm["alpha_peptide_crosslink_position"] == 2
    assert last_csm["alpha_proteins"] == ["Cas10", "Cas11"]
    assert last_csm["alpha_proteins_crosslink_positions"] == [33, 21]
    assert last_csm["alpha_proteins_peptide_positions"] == [32, 20]
    assert last_csm["alpha_score"] == pytest.approx(5.3)
    assert last_csm["alpha_decoy"]
    assert last_csm["beta_peptide"] == "PEKPIDE"
    assert mts(last_csm["beta_modifications"]) == "(3:[DSS|138.06808])"
    assert last_csm["beta_peptide_crosslink_position"] == 3
    assert last_csm["beta_proteins"] == ["Cas9"]
    assert last_csm["beta_proteins_crosslink_positions"] == [28]
    assert last_csm["beta_proteins_peptide_positions"] == [26]
    assert last_csm["beta_score"] == pytest.approx(34.89)
    assert not last_csm["beta_decoy"]
    assert last_csm["crosslink_type"] == "inter"
    assert last_csm["score"] == pytest.approx(5.4)
    assert last_csm["spectrum_file"] == "S1.raw"
    assert last_csm["scan_nr"] == 2
    assert last_csm["charge"] == 4
    assert last_csm["retention_time"] == pytest.approx(37.332)
    assert last_csm["ion_mobility"] == pytest.approx(-70.0)



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


FASTA_SMALL = "data/_fasta/Cas9_plus10.fasta"
FASTA_LARGE = "data/_fasta/uniprotkb_proteome_UP000005640_AND_revi_2025_05_23.fasta"


def build_protein_db(fasta: str):
    from Bio.SeqIO.FastaIO import SimpleFastaParser

    def fasta_title_to_accession(title: str):
        if "|" in title:
            return title.split("|")[1].strip()
        return title.strip()

    protein_db = dict()
    with open(fasta, "r", encoding="utf-8") as f:
        for item in SimpleFastaParser(f):
            protein_db[fasta_title_to_accession(item[0])] = item[1]
    return protein_db


def test1():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    peptide = "ADANLDK"
    with pytest.raises(RuntimeError, match=f"No match found for peptide {peptide}!"):
        _r = __get_proteins_and_positions(peptide, {})


def test2():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    peptide = "ADANLDK"
    with pytest.raises(RuntimeError, match=f"No match found for peptide {peptide}!"):
        _r = __get_proteins_and_positions(peptide, {"PROT1": "PEPKTIDE"})


def test3():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_SMALL)

    peptide = "ADANLDK"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["Cas9"]
    assert prot_pos[1] == [1286]


def test4():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_SMALL)

    peptide = "GNTDRHSIK"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["Cas9"]
    assert prot_pos[1] == [39]


def test5():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_SMALL)

    peptide = "YYRDPCCCRPVSCQTVSRPVTFVPRCTR"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["KRA3_SHEEP"]
    assert prot_pos[1] == [22]


def test6():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_LARGE)

    peptide = "KEESGF"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["A0A087X1C5", "P10635"]
    assert prot_pos[1] == [213, 213]


def test7():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_LARGE)

    peptide = "PGTKLCHGDSELTSGLLAT"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["Q9Y6C7"]
    assert prot_pos[1] == [75]


def test8():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = build_protein_db(FASTA_LARGE)

    peptide = "PGTKLCHGDSELTSGLLAT"
    # this is a performance test to see if it shows up in pytest --durations
    for i in range(1000):
        prot_pos = __get_proteins_and_positions(peptide, protein_db)
        assert prot_pos[0] == ["Q9Y6C7"]
        assert prot_pos[1] == [75]


def test9():
    from pyXLMS.transform import fasta_title_to_accession

    title = "sp|A0A087X1C5|CP2D7_HUMAN Putative cytochrome P450 2D7 OS=Homo sapiens OX=9606 GN=CYP2D7 PE=5 SV=1"
    assert fasta_title_to_accession(title) == "A0A087X1C5"
    assert fasta_title_to_accession("Cas9") == "Cas9"


def test10():
    from pyXLMS.transform import reannotate_positions

    with pytest.raises(
        TypeError,
        match="Can't annotate positions for dict. Dict has to be a valid 'parser_result'!",
    ):
        _r = reannotate_positions({"data_type": "error"}, FASTA_SMALL)


def test11():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_csm_min

    csms = [
        create_csm_min("ADANLDK", 7, "GNTDRHSIK", 9, "RUN_1", 1),
        create_csm_min("GNTDRHSIK", 9, "YYRDPCCCRPVSCQTVSRPVTFVPRCTR", 1, "RUN_1", 2),
    ]
    new_csms = reannotate_positions(csms, FASTA_SMALL)

    assert new_csms[0]["alpha_proteins"] == ["Cas9"]
    assert new_csms[0]["alpha_proteins_crosslink_positions"] == [1293]
    assert new_csms[0]["alpha_proteins_peptide_positions"] == [1287]
    assert new_csms[0]["beta_proteins"] == ["Cas9"]
    assert new_csms[0]["beta_proteins_crosslink_positions"] == [48]
    assert new_csms[0]["beta_proteins_peptide_positions"] == [40]
    assert new_csms[1]["alpha_proteins"] == ["Cas9"]
    assert new_csms[1]["alpha_proteins_crosslink_positions"] == [48]
    assert new_csms[1]["alpha_proteins_peptide_positions"] == [40]
    assert new_csms[1]["beta_proteins"] == ["KRA3_SHEEP"]
    assert new_csms[1]["beta_proteins_crosslink_positions"] == [23]
    assert new_csms[1]["beta_proteins_peptide_positions"] == [23]


def test12():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_crosslink_min

    xls = [
        create_crosslink_min("ADANLDK", 7, "GNTDRHSIK", 9),
        create_crosslink_min("GNTDRHSIK", 9, "YYRDPCCCRPVSCQTVSRPVTFVPRCTR", 1),
    ]
    new_xls = reannotate_positions(xls, FASTA_SMALL)

    assert new_xls[0]["alpha_proteins"] == ["Cas9"]
    assert new_xls[0]["alpha_proteins_crosslink_positions"] == [1293]
    assert new_xls[0]["beta_proteins"] == ["Cas9"]
    assert new_xls[0]["beta_proteins_crosslink_positions"] == [48]
    assert new_xls[1]["alpha_proteins"] == ["Cas9"]
    assert new_xls[1]["alpha_proteins_crosslink_positions"] == [48]
    assert new_xls[1]["beta_proteins"] == ["KRA3_SHEEP"]
    assert new_xls[1]["beta_proteins_crosslink_positions"] == [23]


def test13():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_crosslink_min
    from pyXLMS.data import create_parser_result

    csms = [
        create_csm_min("ADANLDK", 7, "GNTDRHSIK", 9, "RUN_1", 1),
        create_csm_min("GNTDRHSIK", 9, "YYRDPCCCRPVSCQTVSRPVTFVPRCTR", 1, "RUN_1", 2),
    ]
    xls = [
        create_crosslink_min("ADANLDK", 7, "GNTDRHSIK", 9),
        create_crosslink_min("GNTDRHSIK", 9, "YYRDPCCCRPVSCQTVSRPVTFVPRCTR", 1),
    ]
    pr = create_parser_result("TEST", csms=csms, crosslinks=xls)
    new_pr = reannotate_positions(pr, FASTA_SMALL)

    assert new_pr["crosslink-spectrum-matches"][0]["alpha_proteins"] == ["Cas9"]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_crosslink_positions"
    ] == [1293]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_peptide_positions"
    ] == [1287]
    assert new_pr["crosslink-spectrum-matches"][0]["beta_proteins"] == ["Cas9"]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_crosslink_positions"
    ] == [48]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_peptide_positions"
    ] == [40]
    assert new_pr["crosslink-spectrum-matches"][1]["alpha_proteins"] == ["Cas9"]
    assert new_pr["crosslink-spectrum-matches"][1][
        "alpha_proteins_crosslink_positions"
    ] == [48]
    assert new_pr["crosslink-spectrum-matches"][1][
        "alpha_proteins_peptide_positions"
    ] == [40]
    assert new_pr["crosslink-spectrum-matches"][1]["beta_proteins"] == ["KRA3_SHEEP"]
    assert new_pr["crosslink-spectrum-matches"][1][
        "beta_proteins_crosslink_positions"
    ] == [23]
    assert new_pr["crosslink-spectrum-matches"][1][
        "beta_proteins_peptide_positions"
    ] == [23]
    assert new_pr["crosslinks"][0]["alpha_proteins"] == ["Cas9"]
    assert new_pr["crosslinks"][0]["alpha_proteins_crosslink_positions"] == [1293]
    assert new_pr["crosslinks"][0]["beta_proteins"] == ["Cas9"]
    assert new_pr["crosslinks"][0]["beta_proteins_crosslink_positions"] == [48]
    assert new_pr["crosslinks"][1]["alpha_proteins"] == ["Cas9"]
    assert new_pr["crosslinks"][1]["alpha_proteins_crosslink_positions"] == [48]
    assert new_pr["crosslinks"][1]["beta_proteins"] == ["KRA3_SHEEP"]
    assert new_pr["crosslinks"][1]["beta_proteins_crosslink_positions"] == [23]


def test14():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_csm_min

    csms = [create_csm_min("KEESGF", 1, "PGTKLCHGDSELTSGLLAT", 4, "RUN_1", 1)]
    new_csms = reannotate_positions(csms, FASTA_LARGE)

    assert new_csms[0]["alpha_proteins"] == ["A0A087X1C5", "P10635"]
    assert new_csms[0]["alpha_proteins_crosslink_positions"] == [214, 214]
    assert new_csms[0]["alpha_proteins_peptide_positions"] == [214, 214]
    assert new_csms[0]["beta_proteins"] == ["Q9Y6C7"]
    assert new_csms[0]["beta_proteins_crosslink_positions"] == [79]
    assert new_csms[0]["beta_proteins_peptide_positions"] == [76]


def test15():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_crosslink_min

    xls = [create_crosslink_min("KEESGF", 1, "PGTKLCHGDSELTSGLLAT", 4)]
    new_xls = reannotate_positions(xls, FASTA_LARGE)

    assert new_xls[0]["alpha_proteins"] == ["A0A087X1C5", "P10635"]
    assert new_xls[0]["alpha_proteins_crosslink_positions"] == [214, 214]
    assert new_xls[0]["beta_proteins"] == ["Q9Y6C7"]
    assert new_xls[0]["beta_proteins_crosslink_positions"] == [79]


def test16():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_crosslink_min
    from pyXLMS.data import create_parser_result

    csms = [create_csm_min("KEESGF", 1, "PGTKLCHGDSELTSGLLAT", 4, "RUN_1", 1)]
    xls = [create_crosslink_min("KEESGF", 1, "PGTKLCHGDSELTSGLLAT", 4)]
    pr = create_parser_result("TEST", csms=csms, crosslinks=xls)
    new_pr = reannotate_positions(pr, FASTA_LARGE)

    assert new_pr["crosslink-spectrum-matches"][0]["alpha_proteins"] == [
        "A0A087X1C5",
        "P10635",
    ]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_crosslink_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_peptide_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][0]["beta_proteins"] == ["Q9Y6C7"]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_crosslink_positions"
    ] == [79]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_peptide_positions"
    ] == [76]
    assert new_pr["crosslinks"][0]["alpha_proteins"] == ["A0A087X1C5", "P10635"]
    assert new_pr["crosslinks"][0]["alpha_proteins_crosslink_positions"] == [214, 214]
    assert new_pr["crosslinks"][0]["beta_proteins"] == ["Q9Y6C7"]
    assert new_pr["crosslinks"][0]["beta_proteins_crosslink_positions"] == [79]


def test17():
    from pyXLMS.transform import reannotate_positions
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_parser_result

    # this is a performance test to see if it shows up in pytest --durations
    csms = [
        create_csm_min("KEESGF", 1, "PGTKLCHGDSELTSGLLAT", 4, "RUN_1", i + 1)
        for i in range(1000)
    ]
    pr = create_parser_result("TEST", csms=csms, crosslinks=None)
    new_pr = reannotate_positions(pr, FASTA_LARGE)

    assert new_pr["crosslink-spectrum-matches"][0]["alpha_proteins"] == [
        "A0A087X1C5",
        "P10635",
    ]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_crosslink_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][0][
        "alpha_proteins_peptide_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][0]["beta_proteins"] == ["Q9Y6C7"]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_crosslink_positions"
    ] == [79]
    assert new_pr["crosslink-spectrum-matches"][0][
        "beta_proteins_peptide_positions"
    ] == [76]
    assert new_pr["crosslink-spectrum-matches"][-1]["alpha_proteins"] == [
        "A0A087X1C5",
        "P10635",
    ]
    assert new_pr["crosslink-spectrum-matches"][-1][
        "alpha_proteins_crosslink_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][-1][
        "alpha_proteins_peptide_positions"
    ] == [214, 214]
    assert new_pr["crosslink-spectrum-matches"][-1]["beta_proteins"] == ["Q9Y6C7"]
    assert new_pr["crosslink-spectrum-matches"][-1][
        "beta_proteins_crosslink_positions"
    ] == [79]
    assert new_pr["crosslink-spectrum-matches"][-1][
        "beta_proteins_peptide_positions"
    ] == [76]
    assert new_pr["crosslinks"] is None


def test18():
    from pyXLMS.transform.reannotate_positions import __generate_all_sequences
    from pyXLMS.constants import AMINO_ACIDS_REPLACEMENTS

    x = len(AMINO_ACIDS_REPLACEMENTS["X"])
    assert len(set(__generate_all_sequences("PX"))) == x
    assert len(set(__generate_all_sequences("PXX"))) == x * x
    assert len(set(__generate_all_sequences("PXXB"))) == x * x * 2
    assert len(set(__generate_all_sequences("PXXBJ"))) == x * x * 2 * 2
    assert len(set(__generate_all_sequences("PXXBJZ"))) == x * x * 2 * 2 * 2
    assert len(set(__generate_all_sequences("X"))) == x
    assert len(set(__generate_all_sequences("XX"))) == x * x
    assert len(set(__generate_all_sequences("XXB"))) == x * x * 2
    assert len(set(__generate_all_sequences("XXBJ"))) == x * x * 2 * 2
    assert len(set(__generate_all_sequences("XXBJZ"))) == x * x * 2 * 2 * 2
    assert len(set(__generate_all_sequences("PXP"))) == x
    assert len(set(__generate_all_sequences("PXPX"))) == x * x
    assert len(set(__generate_all_sequences("PXXPB"))) == x * x * 2
    assert len(set(__generate_all_sequences("PXXBPJ"))) == x * x * 2 * 2
    assert len(set(__generate_all_sequences("PPXXBJZ"))) == x * x * 2 * 2 * 2


def test19():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = {
        "P76000": (
            "MNTIHLRCLFRMNPLVWCLRADVAAELRSLRRYYHLSNGMESKSVDTRSIYRZLGATLSY"
            "NMRLGNGMEXEPWLKAAVRKEFVDDNRVKVNNDGNFVNDLSGRRGIYQAAIKASFSSTFS"
            "GHLGVGYSHGAGVESPWNAVAGVNWSF"
        )
    }

    peptide = "ELGATLSYNMRLGNGMEKEPWLK"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["P76000"]
    assert prot_pos[1] == [52]


def test20():
    from pyXLMS.transform.reannotate_positions import __get_proteins_and_positions

    protein_db = {
        "P76000": (
            "MNTIHLRCLFRMNPLVWCLRADVAAELRSLRRYYHLSNGMESKSVDTRSIYRZLGATJSY"
            "NMRJGNGMEXEPWLKAAVRKEFVDDNRVKVNNDGNFVNDLSGRRGIYQAAIKASFSSTFS"
            "GHLGVGYSHGAGVESPWNAVAGVNWSF"
        )
    }

    peptide = "ELGATLSYNMRLGNGMEKEPWLK"
    prot_pos = __get_proteins_and_positions(peptide, protein_db)
    assert prot_pos[0] == ["P76000"]
    assert prot_pos[1] == [52]



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

PLINK2_CSMS = "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv"
PLINK3_CSMS = "data/plink3/Cas10_plus10_2025.04.07.filtered_cross-linked_spectra.csv"
PLINK2_XLS = "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv"
PLINK3_XLS = "data/plink3/Cas10_plus10_2025.04.07.filtered_cross-linked_peptides.csv"


def test1():
    from pyXLMS.parser import detect_plink_filetype

    assert (
        detect_plink_filetype(
            "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv"
        )
        == "crosslinks"
    )


def test2():
    from pyXLMS.parser import detect_plink_filetype

    assert (
        detect_plink_filetype(
            "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv"
        )
        == "crosslink-spectrum-matches"
    )


def test3():
    from pyXLMS.parser import detect_plink_filetype

    assert detect_plink_filetype(PLINK2_CSMS) == "crosslink-spectrum-matches"
    assert detect_plink_filetype(PLINK3_CSMS) == "crosslink-spectrum-matches"
    assert detect_plink_filetype(PLINK2_XLS) == "crosslinks"
    assert detect_plink_filetype(PLINK3_XLS) == "crosslinks"
    with open(PLINK2_CSMS, "r", encoding="utf-8") as f:
        assert detect_plink_filetype(f) == "crosslink-spectrum-matches"
    with open(PLINK3_CSMS, "r", encoding="utf-8") as f:
        assert detect_plink_filetype(f) == "crosslink-spectrum-matches"
    with open(PLINK2_XLS, "r", encoding="utf-8") as f:
        assert detect_plink_filetype(f) == "crosslinks"
    with open(PLINK3_XLS, "r", encoding="utf-8") as f:
        assert detect_plink_filetype(f) == "crosslinks"


def test4():
    from pyXLMS.parser import detect_plink_filetype

    with pytest.raises(
        ValueError,
        match="The provided file seems not to be one of the support pLink input files!",
    ):
        _ = detect_plink_filetype(
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.txt"
        )


def test5():
    from pyXLMS.parser.parser_xldbse_plink import (
        __read_plink_cross_linked_peptides_file,
    )

    plink2 = __read_plink_cross_linked_peptides_file(PLINK2_XLS)
    plink3 = __read_plink_cross_linked_peptides_file(PLINK3_XLS)
    assert plink2 is not None
    assert plink3 is not None


def test6():
    from pyXLMS.parser.parser_xldbse_plink import (
        __read_plink_cross_linked_peptides_file,
    )

    with open(PLINK2_XLS, "r", encoding="utf-8") as f:
        plink2 = __read_plink_cross_linked_peptides_file(f)
        f.close()
    with open(PLINK3_XLS, "r", encoding="utf-8") as f:
        plink3 = __read_plink_cross_linked_peptides_file(f)
        f.close()
    assert plink2 is not None
    assert plink3 is not None


def test7():
    from pyXLMS.parser.parser_xldbse_plink import (
        __read_plink_cross_linked_peptides_file,
    )

    with pytest.raises(
        ValueError,
        match="The provided file seems not to be a pLink cross-linked peptides file!",
    ):
        _ = __read_plink_cross_linked_peptides_file(PLINK2_CSMS)


def test8():
    from pyXLMS.parser import read_plink

    crosslinks = read_plink(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_peptides.csv"
    )
    assert len(crosslinks["crosslinks"]) > 0


def test9():
    from pyXLMS.parser import read_plink

    pr = read_plink(PLINK2_XLS)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 252

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "AGFIKR"
    assert xl["alpha_peptide_crosslink_position"] == 5
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [922]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "AGFIKR"
    assert xl["beta_peptide_crosslink_position"] == 5
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [922]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "AGFIKR"
    assert xl["alpha_peptide_crosslink_position"] == 5
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [922]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "GQKNSR"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [779]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LKSVK"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [1162]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "YPKLESEFVYGDYK"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [1007]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None


def test10():
    from pyXLMS.parser import read_plink

    with open(PLINK2_XLS, "r", encoding="utf-8") as f:
        pr = read_plink(f)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 252

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "AGFIKR"
    assert xl["alpha_peptide_crosslink_position"] == 5
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [922]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "AGFIKR"
    assert xl["beta_peptide_crosslink_position"] == 5
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [922]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "AGFIKR"
    assert xl["alpha_peptide_crosslink_position"] == 5
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [922]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "GQKNSR"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [779]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LKSVK"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [1162]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "YPKLESEFVYGDYK"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [1007]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None


def test11():
    from pyXLMS.parser import read_plink

    pr = read_plink(PLINK3_XLS)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 2

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LFKGHPETLEK"
    assert xl["alpha_peptide_crosslink_position"] == 3
    assert xl["alpha_proteins"] == ["sp|MYG_HUMAN|"]
    assert xl["alpha_proteins_crosslink_positions"] == [34]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LFKGHPETLEK"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["sp|MYG_HUMAN|"]
    assert xl["beta_proteins_crosslink_positions"] == [34]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LKYENEMALR"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["sp|K1C15_SHEEP|"]
    assert xl["alpha_proteins_crosslink_positions"] == [192]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LKYENEMALR"
    assert xl["beta_peptide_crosslink_position"] == 2
    assert xl["beta_proteins"] == ["sp|K1C15_SHEEP|"]
    assert xl["beta_proteins_crosslink_positions"] == [192]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None


def test12():
    from pyXLMS.parser import read_plink

    with open(PLINK3_XLS, "r", encoding="utf-8") as f:
        pr = read_plink(f)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 2

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LFKGHPETLEK"
    assert xl["alpha_peptide_crosslink_position"] == 3
    assert xl["alpha_proteins"] == ["sp|MYG_HUMAN|"]
    assert xl["alpha_proteins_crosslink_positions"] == [34]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LFKGHPETLEK"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["sp|MYG_HUMAN|"]
    assert xl["beta_proteins_crosslink_positions"] == [34]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "partial"
    assert xl["alpha_peptide"] == "LKYENEMALR"
    assert xl["alpha_peptide_crosslink_position"] == 2
    assert xl["alpha_proteins"] == ["sp|K1C15_SHEEP|"]
    assert xl["alpha_proteins_crosslink_positions"] == [192]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LKYENEMALR"
    assert xl["beta_peptide_crosslink_position"] == 2
    assert xl["beta_proteins"] == ["sp|K1C15_SHEEP|"]
    assert xl["beta_proteins_crosslink_positions"] == [192]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] is None


def test13():
    from pyXLMS.parser import read_plink
    from pyXLMS.transform import modifications_to_str as mts

    with open(PLINK2_CSMS, "r", encoding="utf-8") as f:
        pr = read_plink(f)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 961

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "FDNLTKAER"
    assert mts(csm["alpha_modifications"]) == "(6:[DSS|138.06808])"
    assert csm["alpha_peptide_crosslink_position"] == 6
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [906]
    assert csm["alpha_proteins_peptide_positions"] == [901]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "YDENDKLIR"
    assert mts(csm["beta_modifications"]) == "(6:[DSS|138.06808])"
    assert csm["beta_peptide_crosslink_position"] == 6
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [952]
    assert csm["beta_proteins_peptide_positions"] == [947]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("5.553153e-009"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 13098
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[118]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "KIECFDSVEISGVEDR"
    assert (
        mts(csm["alpha_modifications"])
        == "(1:[DSS|138.06808]);(4:[Carbamidomethyl|57.021464])"
    )
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [575]
    assert csm["alpha_proteins_peptide_positions"] == [575]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "KIECFDSVEISGVEDR"
    assert (
        mts(csm["beta_modifications"])
        == "(1:[DSS|138.06808]);(4:[Carbamidomethyl|57.021464])"
    )
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [575]
    assert csm["beta_proteins_peptide_positions"] == [575]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("8.010787e-004"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 19140
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[951]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "GKSDNVPSEEVVK"
    assert mts(csm["alpha_modifications"]) == "(2:[DSS|138.06808])"
    assert csm["alpha_peptide_crosslink_position"] == 2
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [870]
    assert csm["alpha_proteins_peptide_positions"] == [869]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "VKYVTEGMR"
    assert (
        mts(csm["beta_modifications"])
        == "(2:[DSS|138.06808]);(8:[Oxidation|15.994915])"
    )
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [532]
    assert csm["beta_proteins_peptide_positions"] == [531]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("4.356499e-001"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 11060
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "LKYENEMALR"
    assert mts(csm["alpha_modifications"]) == "(2:[DSS|138.06808])"
    assert csm["alpha_peptide_crosslink_position"] == 2
    assert csm["alpha_proteins"] == ["sp|K1C15_SHEEP|"]
    assert csm["alpha_proteins_crosslink_positions"] == [192]
    assert csm["alpha_proteins_peptide_positions"] == [191]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "TKYNALK"
    assert mts(csm["beta_modifications"]) == "(2:[DSS|138.06808])"
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["sp|CTRB_BOVIN|"]
    assert csm["beta_proteins_crosslink_positions"] == [145]
    assert csm["beta_proteins_peptide_positions"] == [144]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(float("4.868460e-001"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 11341
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test14():
    from pyXLMS.parser import read_plink
    from pyXLMS.transform import modifications_to_str as mts

    with open(PLINK3_CSMS, "r", encoding="utf-8") as f:
        pr = read_plink(f)
        f.close()

    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 3

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "LFKGHPETLEK"
    assert mts(csm["alpha_modifications"]) == "(3:[DSS|138.06808])"
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] == ["sp|K1C15_SHEEP|", "sp|MYG_HUMAN|"]
    assert csm["alpha_proteins_crosslink_positions"] == [192, 34]
    assert csm["alpha_proteins_peptide_positions"] == [190, 32]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "LKYENEMALR"
    assert (
        mts(csm["beta_modifications"])
        == "(2:[DSS|138.06808]);(7:[Oxidation|15.994915])"
    )
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["sp|K1C15_SHEEP|", "sp|MYG_HUMAN|"]
    assert csm["beta_proteins_crosslink_positions"] == [192, 34]
    assert csm["beta_proteins_peptide_positions"] == [191, 33]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("0.332623"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 14421
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.transform.to_dataframe import __cc

    assert __cc([1, 2, 3]) == "1;2;3"


def test6():
    from pyXLMS.transform.to_dataframe import __cc

    assert __cc([1, 2, 3], ",") == "1,2,3"


def test7():
    from pyXLMS.transform.to_dataframe import __cc

    assert __cc([]) == ""


def test8():
    from pyXLMS.transform.to_dataframe import __cc

    assert __cc(None) is None


def test9():
    from pyXLMS import data
    from pyXLMS.transform.to_dataframe import __crosslinks_to_dataframe

    c1 = data.create_crosslink(
        "PEPTIDE",
        3,
        ["PROTEINA"],
        [5],
        False,
        "PEPTIDE"[::-1],
        5,
        ["PROTEINA"],
        [5],
        True,
        70.3,
    )
    c2 = data.create_crosslink(
        "PEPTIDEB",
        3,
        ["PROTEINB"],
        [5],
        False,
        "PEPTIDEA",
        5,
        ["PROTEINA", "PROTEINC"],
        [1, 2],
        False,
        123.7,
    )
    crosslinks = [c1, c2]
    df = __crosslinks_to_dataframe(crosslinks)

    assert df.shape[0] == 2
    assert df.shape[1] == 13
    assert df.loc[0, "Completeness"] == "full"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[0, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[0, "Alpha Proteins Crosslink Positions"] == "5"
    # assert df.loc[0, "Alpha Decoy"] == True
    assert df.loc[0, "Alpha Decoy"]
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[0, "Beta Proteins"] == "PROTEINA"
    assert df.loc[0, "Beta Proteins Crosslink Positions"] == "5"
    # assert df.loc[0, "Beta Decoy"] == False
    assert not df.loc[0, "Beta Decoy"]
    assert df.loc[0, "Crosslink Type"] == "intra"
    assert df.loc[0, "Crosslink Score"] == pytest.approx(70.3)
    assert df.loc[1, "Completeness"] == "full"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[1, "Alpha Proteins"] == "PROTEINA;PROTEINC"
    assert df.loc[1, "Alpha Proteins Crosslink Positions"] == "1;2"
    # assert df.loc[1, "Alpha Decoy"] == False
    assert not df.loc[1, "Alpha Decoy"]
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[1, "Beta Proteins"] == "PROTEINB"
    assert df.loc[1, "Beta Proteins Crosslink Positions"] == "5"
    # assert df.loc[1, "Beta Decoy"] == False
    assert not df.loc[1, "Beta Decoy"]
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert df.loc[1, "Crosslink Score"] == pytest.approx(123.7)


def test10():
    from pyXLMS import data, transform

    c1 = data.create_crosslink(
        "PEPTIDE",
        3,
        ["PROTEINA"],
        [5],
        False,
        "PEPTIDE"[::-1],
        5,
        ["PROTEINA"],
        [5],
        True,
        70.3,
    )
    c2 = data.create_crosslink(
        "PEPTIDEB",
        3,
        ["PROTEINB"],
        [5],
        False,
        "PEPTIDEA",
        5,
        ["PROTEINA", "PROTEINC"],
        [1, 2],
        False,
        123.7,
    )
    crosslinks = [c1, c2]
    df = transform.to_dataframe(crosslinks)

    assert df.shape[0] == 2
    assert df.shape[1] == 13
    assert df.loc[0, "Completeness"] == "full"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[0, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[0, "Alpha Proteins Crosslink Positions"] == "5"
    # assert df.loc[0, "Alpha Decoy"] == True
    assert df.loc[0, "Alpha Decoy"]
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[0, "Beta Proteins"] == "PROTEINA"
    assert df.loc[0, "Beta Proteins Crosslink Positions"] == "5"
    # assert df.loc[0, "Beta Decoy"] == False
    assert not df.loc[0, "Beta Decoy"]
    assert df.loc[0, "Crosslink Type"] == "intra"
    assert df.loc[0, "Crosslink Score"] == pytest.approx(70.3)
    assert df.loc[1, "Completeness"] == "full"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[1, "Alpha Proteins"] == "PROTEINA;PROTEINC"
    assert df.loc[1, "Alpha Proteins Crosslink Positions"] == "1;2"
    # assert df.loc[1, "Alpha Decoy"] == False
    assert not df.loc[1, "Alpha Decoy"]
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[1, "Beta Proteins"] == "PROTEINB"
    assert df.loc[1, "Beta Proteins Crosslink Positions"] == "5"
    # assert df.loc[1, "Beta Decoy"] == False
    assert not df.loc[1, "Beta Decoy"]
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert df.loc[1, "Crosslink Score"] == pytest.approx(123.7)


def test11():
    from pyXLMS import data, transform
    import pandas as pd

    c1 = data.create_crosslink(
        "PEPTIDE",
        3,
        None,
        None,
        None,
        "PEPTIDE"[::-1],
        5,
        None,
        None,
        None,
        None,
    )
    c2 = data.create_crosslink(
        "PEPTIDEB",
        3,
        None,
        None,
        None,
        "PEPTIDEA",
        5,
        None,
        None,
        None,
        None,
    )
    crosslinks = [c1, c2]
    df = transform.to_dataframe(crosslinks)

    assert df.shape[0] == 2
    assert df.shape[1] == 13
    assert df.loc[0, "Completeness"] == "partial"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert pd.isna(df.loc[0, "Alpha Proteins"])
    assert pd.isna(df.loc[0, "Alpha Proteins Crosslink Positions"])
    assert pd.isna(df.loc[0, "Alpha Decoy"])
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert pd.isna(df.loc[0, "Beta Proteins"])
    assert pd.isna(df.loc[0, "Beta Proteins Crosslink Positions"])
    assert pd.isna(df.loc[0, "Beta Decoy"])
    assert df.loc[0, "Crosslink Type"] == "inter"
    assert pd.isna(df.loc[0, "Crosslink Score"])
    assert df.loc[1, "Completeness"] == "partial"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert pd.isna(df.loc[1, "Alpha Proteins"])
    assert pd.isna(df.loc[1, "Alpha Proteins Crosslink Positions"])
    assert pd.isna(df.loc[1, "Alpha Decoy"])
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert pd.isna(df.loc[1, "Beta Proteins"])
    assert pd.isna(df.loc[1, "Beta Proteins Crosslink Positions"])
    assert pd.isna(df.loc[1, "Beta Decoy"])
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert pd.isna(df.loc[1, "Crosslink Score"])


def test12():
    from pyXLMS import data
    from pyXLMS.transform.to_dataframe import __csms_to_dataframe

    c1 = data.create_csm(
        "PEPTIDE",
        {1: ("Oxidation", 15.994915), 5: ("Carbamidomethyl", 57.021464)},
        3,
        ["PROTEINA"],
        [3],
        [1],
        70.3,
        False,
        "PEPTIDE"[::-1],
        {1: ("Oxidation", 15.994915)},
        5,
        ["PROTEINA"],
        [6],
        [2],
        20.4,
        True,
        score=70.3,
        spectrum_file="MS_EXP1",
        scan_nr=1,
        charge=4,
        rt=12.8,
        im_cv=-50.0,
    )
    c2 = data.create_csm(
        "PEPTIDEB",
        {1: ("Oxidation", 15.994915), 5: ("Carbamidomethyl", 57.021464)},
        3,
        ["PROTEINB", "PROTEINC"],
        [3, 4],
        [1, 2],
        71.3,
        False,
        "PEPTIDEA",
        {},
        5,
        ["PROTEINA"],
        [6],
        [2],
        21.4,
        False,
        score=71.3,
        spectrum_file="MS_EXP1",
        scan_nr=2,
        charge=3,
        rt=12.9,
        im_cv=-70.0,
    )
    csms = [c1, c2]
    df = __csms_to_dataframe(csms)

    assert df.shape[0] == 2
    assert df.shape[1] == 24
    assert df.loc[0, "Completeness"] == "full"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert df.loc[0, "Alpha Peptide Modifications"] == "(1:[Oxidation|15.994915])"
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[0, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[0, "Alpha Proteins Crosslink Positions"] == "6"
    assert df.loc[0, "Alpha Proteins Peptide Positions"] == "2"
    assert df.loc[0, "Alpha Score"] == pytest.approx(20.4)
    # assert df.loc[0, "Alpha Decoy"] == True
    assert df.loc[0, "Alpha Decoy"]
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert (
        df.loc[0, "Beta Peptide Modifications"]
        == "(1:[Oxidation|15.994915]);(5:[Carbamidomethyl|57.021464])"
    )
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[0, "Beta Proteins"] == "PROTEINA"
    assert df.loc[0, "Beta Proteins Crosslink Positions"] == "3"
    assert df.loc[0, "Beta Proteins Peptide Positions"] == "1"
    assert df.loc[0, "Beta Score"] == pytest.approx(70.3)
    # assert df.loc[0, "Beta Decoy"] == False
    assert not df.loc[0, "Beta Decoy"]
    assert df.loc[0, "Crosslink Type"] == "intra"
    assert df.loc[0, "CSM Score"] == pytest.approx(70.3)
    assert df.loc[0, "Spectrum File"] == "MS_EXP1"
    assert df.loc[0, "Scan Nr"] == 1
    assert df.loc[0, "Precursor Charge"] == 4
    assert df.loc[0, "Retention Time"] == pytest.approx(12.8)
    assert df.loc[0, "Ion Mobility"] == pytest.approx(-50)
    assert df.loc[1, "Completeness"] == "full"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert df.loc[1, "Alpha Peptide Modifications"] == ""
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[1, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[1, "Alpha Proteins Crosslink Positions"] == "6"
    assert df.loc[1, "Alpha Proteins Peptide Positions"] == "2"
    assert df.loc[1, "Alpha Score"] == pytest.approx(21.4)
    # assert df.loc[1, "Alpha Decoy"] == False
    assert not df.loc[1, "Alpha Decoy"]
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert (
        df.loc[1, "Beta Peptide Modifications"]
        == "(1:[Oxidation|15.994915]);(5:[Carbamidomethyl|57.021464])"
    )
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[1, "Beta Proteins"] == "PROTEINB;PROTEINC"
    assert df.loc[1, "Beta Proteins Crosslink Positions"] == "3;4"
    assert df.loc[1, "Beta Proteins Peptide Positions"] == "1;2"
    assert df.loc[1, "Beta Score"] == pytest.approx(71.3)
    # assert df.loc[1, "Beta Decoy"] == False
    assert not df.loc[1, "Beta Decoy"]
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert df.loc[1, "CSM Score"] == pytest.approx(71.3)
    assert df.loc[1, "Spectrum File"] == "MS_EXP1"
    assert df.loc[1, "Scan Nr"] == 2
    assert df.loc[1, "Precursor Charge"] == 3
    assert df.loc[1, "Retention Time"] == pytest.approx(12.9)
    assert df.loc[1, "Ion Mobility"] == pytest.approx(-70)


def test13():
    from pyXLMS import data, transform

    c1 = data.create_csm(
        "PEPTIDE",
        {1: ("Oxidation", 15.994915), 5: ("Carbamidomethyl", 57.021464)},
        3,
        ["PROTEINA"],
        [3],
        [1],
        70.3,
        False,
        "PEPTIDE"[::-1],
        {1: ("Oxidation", 15.994915)},
        5,
        ["PROTEINA"],
        [6],
        [2],
        20.4,
        True,
        score=70.3,
        spectrum_file="MS_EXP1",
        scan_nr=1,
        charge=4,
        rt=12.8,
        im_cv=-50.0,
    )
    c2 = data.create_csm(
        "PEPTIDEB",
        {1: ("Oxidation", 15.994915), 5: ("Carbamidomethyl", 57.021464)},
        3,
        ["PROTEINB", "PROTEINC"],
        [3, 4],
        [1, 2],
        71.3,
        False,
        "PEPTIDEA",
        {},
        5,
        ["PROTEINA"],
        [6],
        [2],
        21.4,
        False,
        score=71.3,
        spectrum_file="MS_EXP1",
        scan_nr=2,
        charge=3,
        rt=12.9,
        im_cv=-70.0,
    )
    csms = [c1, c2]
    df = transform.to_dataframe(csms)

    assert df.shape[0] == 2
    assert df.shape[1] == 24
    assert df.loc[0, "Completeness"] == "full"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert df.loc[0, "Alpha Peptide Modifications"] == "(1:[Oxidation|15.994915])"
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[0, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[0, "Alpha Proteins Crosslink Positions"] == "6"
    assert df.loc[0, "Alpha Proteins Peptide Positions"] == "2"
    assert df.loc[0, "Alpha Score"] == pytest.approx(20.4)
    # assert df.loc[0, "Alpha Decoy"] == True
    assert df.loc[0, "Alpha Decoy"]
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert (
        df.loc[0, "Beta Peptide Modifications"]
        == "(1:[Oxidation|15.994915]);(5:[Carbamidomethyl|57.021464])"
    )
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[0, "Beta Proteins"] == "PROTEINA"
    assert df.loc[0, "Beta Proteins Crosslink Positions"] == "3"
    assert df.loc[0, "Beta Proteins Peptide Positions"] == "1"
    assert df.loc[0, "Beta Score"] == pytest.approx(70.3)
    # assert df.loc[0, "Beta Decoy"] == False
    assert not df.loc[0, "Beta Decoy"]
    assert df.loc[0, "Crosslink Type"] == "intra"
    assert df.loc[0, "CSM Score"] == pytest.approx(70.3)
    assert df.loc[0, "Spectrum File"] == "MS_EXP1"
    assert df.loc[0, "Scan Nr"] == 1
    assert df.loc[0, "Precursor Charge"] == 4
    assert df.loc[0, "Retention Time"] == pytest.approx(12.8)
    assert df.loc[0, "Ion Mobility"] == pytest.approx(-50)
    assert df.loc[1, "Completeness"] == "full"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert df.loc[1, "Alpha Peptide Modifications"] == ""
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert df.loc[1, "Alpha Proteins"] == "PROTEINA"
    assert df.loc[1, "Alpha Proteins Crosslink Positions"] == "6"
    assert df.loc[1, "Alpha Proteins Peptide Positions"] == "2"
    assert df.loc[1, "Alpha Score"] == pytest.approx(21.4)
    # assert df.loc[1, "Alpha Decoy"] == False
    assert not df.loc[1, "Alpha Decoy"]
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert (
        df.loc[1, "Beta Peptide Modifications"]
        == "(1:[Oxidation|15.994915]);(5:[Carbamidomethyl|57.021464])"
    )
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert df.loc[1, "Beta Proteins"] == "PROTEINB;PROTEINC"
    assert df.loc[1, "Beta Proteins Crosslink Positions"] == "3;4"
    assert df.loc[1, "Beta Proteins Peptide Positions"] == "1;2"
    assert df.loc[1, "Beta Score"] == pytest.approx(71.3)
    # assert df.loc[1, "Beta Decoy"] == False
    assert not df.loc[1, "Beta Decoy"]
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert df.loc[1, "CSM Score"] == pytest.approx(71.3)
    assert df.loc[1, "Spectrum File"] == "MS_EXP1"
    assert df.loc[1, "Scan Nr"] == 2
    assert df.loc[1, "Precursor Charge"] == 3
    assert df.loc[1, "Retention Time"] == pytest.approx(12.9)
    assert df.loc[1, "Ion Mobility"] == pytest.approx(-70.0)


def test14():
    from pyXLMS import data, transform
    import pandas as pd

    c1 = data.create_csm(
        "PEPTIDE",
        None,
        3,
        None,
        None,
        None,
        None,
        None,
        "PEPTIDE"[::-1],
        None,
        5,
        None,
        None,
        None,
        None,
        None,
        score=None,
        spectrum_file="MS_EXP1",
        scan_nr=1,
        charge=None,
        rt=None,
        im_cv=None,
    )
    c2 = data.create_csm(
        "PEPTIDEB",
        None,
        3,
        None,
        None,
        None,
        None,
        None,
        "PEPTIDEA",
        None,
        5,
        None,
        None,
        None,
        None,
        None,
        score=None,
        spectrum_file="MS_EXP1",
        scan_nr=2,
        charge=None,
        rt=None,
        im_cv=None,
    )
    csms = [c1, c2]
    df = transform.to_dataframe(csms)

    assert df.shape[0] == 2
    assert df.shape[1] == 24
    assert df.loc[0, "Completeness"] == "partial"
    assert df.loc[0, "Alpha Peptide"] == "PEPTIDE"[::-1]
    assert pd.isna(df.loc[0, "Alpha Peptide Modifications"])
    assert df.loc[0, "Alpha Peptide Crosslink Position"] == 5
    assert pd.isna(df.loc[0, "Alpha Proteins"])
    assert pd.isna(df.loc[0, "Alpha Proteins Crosslink Positions"])
    assert pd.isna(df.loc[0, "Alpha Proteins Peptide Positions"])
    assert pd.isna(df.loc[0, "Alpha Score"])
    assert pd.isna(df.loc[0, "Alpha Decoy"])
    assert df.loc[0, "Beta Peptide"] == "PEPTIDE"
    assert pd.isna(df.loc[0, "Beta Peptide Modifications"])
    assert df.loc[0, "Beta Peptide Crosslink Position"] == 3
    assert pd.isna(df.loc[0, "Beta Proteins"])
    assert pd.isna(df.loc[0, "Beta Proteins Crosslink Positions"])
    assert pd.isna(df.loc[0, "Beta Proteins Peptide Positions"])
    assert pd.isna(df.loc[0, "Beta Score"])
    assert pd.isna(df.loc[0, "Beta Decoy"])
    assert df.loc[0, "Crosslink Type"] == "inter"
    assert pd.isna(df.loc[0, "CSM Score"])
    assert df.loc[0, "Spectrum File"] == "MS_EXP1"
    assert df.loc[0, "Scan Nr"] == 1
    assert pd.isna(df.loc[0, "Precursor Charge"])
    assert pd.isna(df.loc[0, "Retention Time"])
    assert pd.isna(df.loc[0, "Ion Mobility"])
    assert df.loc[1, "Completeness"] == "partial"
    assert df.loc[1, "Alpha Peptide"] == "PEPTIDEA"
    assert pd.isna(df.loc[1, "Alpha Peptide Modifications"])
    assert df.loc[1, "Alpha Peptide Crosslink Position"] == 5
    assert pd.isna(df.loc[1, "Alpha Proteins"])
    assert pd.isna(df.loc[1, "Alpha Proteins Crosslink Positions"])
    assert pd.isna(df.loc[1, "Alpha Proteins Peptide Positions"])
    assert pd.isna(df.loc[1, "Alpha Score"])
    assert pd.isna(df.loc[1, "Alpha Decoy"])
    assert df.loc[1, "Beta Peptide"] == "PEPTIDEB"
    assert pd.isna(df.loc[1, "Beta Peptide Modifications"])
    assert df.loc[1, "Beta Peptide Crosslink Position"] == 3
    assert pd.isna(df.loc[1, "Beta Proteins"])
    assert pd.isna(df.loc[1, "Beta Proteins Crosslink Positions"])
    assert pd.isna(df.loc[1, "Beta Proteins Peptide Positions"])
    assert pd.isna(df.loc[1, "Beta Score"])
    assert pd.isna(df.loc[1, "Beta Decoy"])
    assert df.loc[1, "Crosslink Type"] == "inter"
    assert pd.isna(df.loc[1, "CSM Score"])
    assert df.loc[1, "Spectrum File"] == "MS_EXP1"
    assert df.loc[1, "Scan Nr"] == 2
    assert pd.isna(df.loc[1, "Precursor Charge"])
    assert pd.isna(df.loc[1, "Retention Time"])
    assert pd.isna(df.loc[1, "Ion Mobility"])


def test15():
    from pyXLMS import transform

    data = [{"data_type": "peptide-spectrum-match"}]
    with pytest.raises(TypeError, match="The given data object is not supported!"):
        _df = transform.to_dataframe(data)


def test16():
    from pyXLMS import transform

    data = [{"data-type": "peptide-spectrum-match"}]
    with pytest.raises(TypeError, match="The given data object is not supported!"):
        _df = transform.to_dataframe(data)


def test17():
    from pyXLMS import transform

    data = []
    with pytest.raises(
        ValueError, match="Parameter data has to be at least of length one!"
    ):
        _df = transform.to_dataframe(data)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings
import pandas as pd
from tqdm import tqdm

from ..data import check_input
from ..data import create_crosslink
from ..data import create_csm
from ..data import create_parser_result
from ..constants import SCOUT_MODIFICATION_MAPPING
from .util import format_sequence
from .util import get_bool_from_value
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def detect_scout_filetype(
    data: pd.DataFrame,
) -> Literal["scout_csms_unfiltered", "scout_csms_filtered", "scout_xl"]:
    r"""Detects the Scout-related source of the data.

    Detects whether the input data is unfiltered crosslink-spectrum-matches, filtered crosslink-spectrum-matches,
    or crosslinks from Scout.

    Parameters
    ----------
    data : pd.DataFrame
        The input data originating from Scout.

    Returns
    -------
    str
        "scout_csms_unfiltered" if a Scout unfiltered CSMs file was read, "scout_csms_filtered" if a Scout filtered CSMs file was read,
        "scout_xl" if a Scout crosslink/residue pair result file was read.

    Raises
    ------
    ValueError
        If the data source could not be determined.

    Examples
    --------
    >>> from pyXLMS.parser import detect_scout_filetype
    >>> import pandas as pd
    >>> df1 = pd.read_csv("data/scout/Cas9_Unfiltered_CSMs.csv")
    >>> detect_scout_filetype(df1)
    'scout_csms_unfiltered'

    >>> from pyXLMS.parser import detect_scout_filetype
    >>> import pandas as pd
    >>> df2 = pd.read_csv("data/scout/Cas9_Filtered_CSMs.csv")
    >>> detect_scout_filetype(df2)
    'scout_csms_filtered'

    >>> from pyXLMS.parser import detect_scout_filetype
    >>> import pandas as pd
    >>> df3 = pd.read_csv("data/scout/Cas9_Residue_Pairs.csv")
    >>> detect_scout_filetype(df3)
    'scout_xl'
    """
    ## check input
    _ok = check_input(data, "data", pd.DataFrame)

    col_names = data.columns.values.tolist()
    if "ScanNumber" in col_names:
        return "scout_csms_unfiltered"
    if "Scan" in col_names:
        return "scout_csms_filtered"
    if "CSM count" in col_names:
        return "scout_xl"

    raise ValueError(
        "Could not infer data source, are you sure you read a Scout result file?"
    )

    return "err"


def parse_modifications_from_scout_sequence(
    seq: str,
    crosslink_position: int,
    crosslinker: str,
    crosslinker_mass: float,
    modifications: Dict[str, Tuple[str, float]] = SCOUT_MODIFICATION_MAPPING,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[int, Tuple[str, float]]:
    r"""Parse post-translational-modifications from a Scout peptide sequence.

    Parses post-translational-modifications (PTMs) from a Scout peptide sequence,
    for example "M(+15.994900)LASAGELQKGNELALPSK".

    Parameters
    ----------
    seq : str
        The Scout sequence string.
    crosslink_position : int
        Position of the crosslinker in the sequence (1-based).
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    modifications: dict of str, float, default = ``constants.SCOUT_MODIFICATION_MAPPING``
        Mapping of modification names to modification masses.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict of int, tuple
        The ``pyXLMS`` specific modifications object, a dictionary that maps positions to their corresponding modifications and their
        monoisotopic masses.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed (only if ``verbose = 2``).
    KeyError
        If an unknown modification is encountered.

    Examples
    --------
    >>> from pyXLMS.parser import parse_modifications_from_scout_sequence
    >>> seq = "M(+15.994900)LASAGELQKGNELALPSK"
    >>> parse_modifications_from_scout_sequence(seq, 10, "DSS", 138.06808)
    {10: ('DSS', 138.06808), 1: ('Oxidation', 15.994915)}

    >>> from pyXLMS.parser import parse_modifications_from_scout_sequence
    >>> seq = "KIEC(+57.021460)FDSVEISGVEDR"
    >>> parse_modifications_from_scout_sequence(seq, 1, "DSS", 138.06808)
    {1: ('DSS', 138.06808), 4: ('Carbamidomethyl', 57.021464)}
    """
    # clean seq
    sequence = seq.strip()
    # init parsed modifications dict
    parsed_modifications = {crosslink_position: (crosslinker, crosslinker_mass)}
    # parse modifications from sequence
    pos = 0
    current_mod = ""
    for i, aa in enumerate(sequence):
        if aa.isupper():
            pos += 1
            current_mod = ""
        else:
            current_mod += aa
            if (i + 1 >= len(sequence)) or (sequence[i + 1].isupper()):
                mod_key = current_mod.strip("()").strip()
                if mod_key not in modifications:
                    raise KeyError(
                        f"Key {mod_key} not found in parameter 'modifications'. Are you missing a modification?"
                    )
                if pos in parsed_modifications:
                    err_str = (
                        f"Modification at position {pos} already exists!\n"
                        f"Sequence: {sequence}, Crosslink position: {crosslink_position}"
                    )
                    if verbose == 1:
                        warnings.warn(RuntimeWarning(err_str))
                    elif verbose == 2:
                        raise RuntimeError(err_str)
                    t1 = parsed_modifications[pos][0] + "," + modifications[mod_key][0]
                    t2 = parsed_modifications[pos][1] + modifications[mod_key][1]
                    parsed_modifications[pos] = (t1, t2)
                else:
                    parsed_modifications[pos] = modifications[mod_key]
    return parsed_modifications


def __read_scout_csms_unfiltered(
    data: pd.DataFrame,
    crosslinker: str,
    crosslinker_mass: float,
    parse_modifications: bool,
    modifications: Dict[str, Tuple[str, float]],
    verbose: Literal[0, 1, 2],
) -> List[Dict[str, Any]]:
    r"""Reads crosslink-spectrum-matches from a Scout unfiltered CSMs result.

    Parameters
    ----------
    data : pandas.DataFrame
        The Scout unfiltered CSMs result data.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    parse_modifications : bool
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple
        Mapping of Scout sequence elements (e.g. ``"+15.994900"``) and modifications (e.g ``"Oxidation of Methionine"``)
        to their modifications (e.g. ``("Oxidation", 15.994915)``).
    verbose : 0, 1, or 2
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    list of dict
        The read crosslink-spectrum-matches.

    Notes
    -----
    This function should not be called directly, it is called from ``read_scout()``.
    """
    csms = list()
    xl = data.dropna(axis=0, subset=["AlphaPeptide", "BetaPeptide"])
    for i, row in tqdm(
        xl.iterrows(), total=xl.shape[0], desc="Reading Scout unfiltered CSMs..."
    ):
        csm = create_csm(
            peptide_a=format_sequence(str(row["AlphaPeptide"])),
            modifications_a=parse_modifications_from_scout_sequence(
                str(row["AlphaPeptide"]),
                int(row["AlphaPos"]) + 1,
                crosslinker,
                crosslinker_mass,
                modifications,
                verbose,
            )
            if parse_modifications
            else None,
            xl_position_peptide_a=int(row["AlphaPos"]) + 1,
            proteins_a=[
                protein.strip() for protein in str(row["AlphaMappings"]).split(";")
            ],
            xl_position_proteins_a=None,
            pep_position_proteins_a=None,
            score_a=float(row["AlphaScore"]),
            decoy_a=str(row["Class"]).strip() in ["FullDecoy", "BetaTarget"],
            peptide_b=format_sequence(str(row["BetaPeptide"])),
            modifications_b=parse_modifications_from_scout_sequence(
                str(row["BetaPeptide"]),
                int(row["BetaPos"]) + 1,
                crosslinker,
                crosslinker_mass,
                modifications,
                verbose,
            )
            if parse_modifications
            else None,
            xl_position_peptide_b=int(row["BetaPos"]) + 1,
            proteins_b=[
                protein.strip() for protein in str(row["BetaMappings"]).split(";")
            ],
            xl_position_proteins_b=None,
            pep_position_proteins_b=None,
            score_b=float(row["BetaScore"]),
            decoy_b=str(row["Class"]) in ["FullDecoy", "AlphaTarget"],
            score=float(row["XLScore"]),
            spectrum_file=str(row["FileName"]).strip(),
            scan_nr=int(row["ScanNumber"]),
            charge=int(row["Charge"]),
            rt=None,
            im_cv=None,
            additional_information={
                "source": __serialize_pandas_series(row),
                "ClassificationScore": float(row["ClassificationScore"])
                if "ClassificationScore" in row.index
                else None,
                "XlinkxAlpha": float(row["XlinkxAlpha"])
                if "XlinkxAlpha" in row.index
                else None,
                "XlinkxBeta": float(row["XlinkxBeta"])
                if "XlinkxBeta" in row.index
                else None,
                "XlinkxScore": float(row["XlinkxScore"])
                if "XlinkxScore" in row.index
                else None,
                "PoissonScore": float(row["PoissonScore"])
                if "PoissonScore" in row.index
                else None,
            },
        )
        csms.append(csm)
    return csms


def __read_scout_csms_filtered(
    data: pd.DataFrame,
    crosslinker: str,
    crosslinker_mass: float,
    parse_modifications: bool,
    modifications: Dict[str, Tuple[str, float]],
    verbose: Literal[0, 1, 2],
) -> List[Dict[str, Any]]:
    r"""Reads crosslink-spectrum-matches from a Scout filtered CSMs result.

    Parameters
    ----------
    data : pandas.DataFrame
        The Scout filtered CSMs result data.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float
        Monoisotopic delta mass of the crosslink modification.
    parse_modifications : bool
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple
        Mapping of Scout sequence elements (e.g. ``"+15.994900"``) and modifications (e.g ``"Oxidation of Methionine"``)
        to their modifications (e.g. ``("Oxidation", 15.994915)``).
    verbose : 0, 1, or 2
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    list of dict
        The read crosslink-spectrum-matches.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed (only if ``verbose = 2``).
    KeyError
        If an unknown modification is encountered.

    Notes
    -----
    This function should not be called directly, it is called from ``read_scout()``.
    """

    ## helper functions
    def str_contains(s: str, contains: List[str]) -> bool:
        for subs in contains:
            if subs in s:
                return True
        return False

    def parse_modifications_fn(
        row: pd.Series,
        alpha: bool,
        crosslinker: str,
        crosslinker_mass: float,
        modifications: Dict[str, Tuple[str, float]] = SCOUT_MODIFICATION_MAPPING,
        verbose: Literal[0, 1, 2] = 1,
    ) -> Dict[int, Tuple[str, float]]:
        sequence = (
            str(row["Alpha peptide"]).strip()
            if alpha
            else str(row["Beta peptide"]).strip()
        )
        crosslink_position = (
            int(row["Alpha peptide position"])
            if alpha
            else int(row["Beta peptide position"])
        )
        if alpha and "Alpha modification(s)" not in row.index:
            return parse_modifications_from_scout_sequence(
                seq=str(row["Modified alpha peptide"]),
                crosslink_position=crosslink_position,
                crosslinker=crosslinker,
                crosslinker_mass=crosslinker_mass,
                modifications=modifications,
                verbose=verbose,
            )
        if not alpha and "Beta modification(s)" not in row.index:
            return parse_modifications_from_scout_sequence(
                seq=str(row["Modified beta peptide"]),
                crosslink_position=crosslink_position,
                crosslinker=crosslinker,
                crosslinker_mass=crosslinker_mass,
                modifications=modifications,
                verbose=verbose,
            )
        parsed_modifications = {crosslink_position: (crosslinker, crosslinker_mass)}
        if alpha and bool(pd.isna(row["Alpha modification(s)"])):
            return parsed_modifications
        if not alpha and bool(pd.isna(row["Beta modification(s)"])):
            return parsed_modifications
        mods = (
            str(row["Alpha modification(s)"]).split(";")
            if alpha
            else str(row["Beta modification(s)"]).split(";")
        )
        for mod in mods:
            rpos = mod.split("(")[0].strip()
            mod_key = mod.split("(")[1].rstrip(")").strip()
            pos = -1
            if str_contains(
                rpos.lower(),
                [
                    "nterm",
                    "nterminal",
                    "nterminus",
                    "n-term",
                    "n-terminal",
                    "n-terminus",
                ],
            ):
                pos = 0
            elif str_contains(
                rpos.lower(),
                [
                    "cterm",
                    "cterminal",
                    "cterminus",
                    "c-term",
                    "c-terminal",
                    "c-terminus",
                ],
            ):
                pos = len(sequence)
            else:
                pos = int(rpos[1:])
            if mod_key not in modifications:
                raise KeyError(
                    f"Key {mod_key} not found in parameter 'modifications'. Are you missing a modification?"
                )
            if pos in parsed_modifications:
                err_str = (
                    f"Modification at position {pos} already exists!\n"
                    f"CSM Scan Number: {int(row['Scan'])}!\n"
                    f"Sequence: {sequence}, Crosslink position: {crosslink_position}, Modifications: {';'.join(mods)}"
                )
                if verbose == 1:
                    warnings.warn(RuntimeWarning(err_str))
                elif verbose == 2:
                    raise RuntimeError(err_str)
                t1 = parsed_modifications[pos][0] + "," + modifications[mod_key][0]
                t2 = parsed_modifications[pos][1] + modifications[mod_key][1]
                parsed_modifications[pos] = (t1, t2)
            else:
                parsed_modifications[pos] = modifications[mod_key]
        return parsed_modifications

    ## create csms
    csms = list()
    xl = data.dropna(axis=0, subset=["Alpha peptide", "Beta peptide"])
    for i, row in tqdm(
        xl.iterrows(), total=xl.shape[0], desc="Reading Scout filtered CSMs..."
    ):
        csm = create_csm(
            peptide_a=format_sequence(str(row["Alpha peptide"])),
            modifications_a=parse_modifications_fn(
                row,
                True,
                crosslinker,
                crosslinker_mass,
                modifications,
                verbose,
            )
            if parse_modifications
            else None,
            xl_position_peptide_a=int(row["Alpha peptide position"]),
            proteins_a=[
                protein.strip()
                for protein in str(row["Alpha protein mapping(s)"]).split(";")
            ],
            xl_position_proteins_a=[
                int(pos) for pos in str(row["Alpha protein(s) position(s)"]).split(";")
            ],
            pep_position_proteins_a=[
                int(pos) - int(row["Alpha peptide position"]) + 1
                for pos in str(row["Alpha protein(s) position(s)"]).split(";")
            ],
            score_a=None,
            decoy_a=get_bool_from_value(row["IsDecoy"]),
            peptide_b=format_sequence(str(row["Beta peptide"])),
            modifications_b=parse_modifications_fn(
                row,
                False,
                crosslinker,
                crosslinker_mass,
                modifications,
                verbose,
            )
            if parse_modifications
            else None,
            xl_position_peptide_b=int(row["Beta peptide position"]),
            proteins_b=[
                protein.strip()
                for protein in str(row["Beta protein mapping(s)"]).split(";")
            ],
            xl_position_proteins_b=[
                int(pos) for pos in str(row["Beta protein(s) position(s)"]).split(";")
            ],
            pep_position_proteins_b=[
                int(pos) - int(row["Beta peptide position"]) + 1
                for pos in str(row["Beta protein(s) position(s)"]).split(";")
            ],
            score_b=None,
            decoy_b=get_bool_from_value(row["IsDecoy"]),
            score=float(row["Score"]),
            spectrum_file=str(row["File"]).strip(),
            scan_nr=int(row["Scan"]),
            charge=int(row["Precursor charge"]),
            rt=None,
            im_cv=None,
            additional_information={"source": __serialize_pandas_series(row)},
        )
        csms.append(csm)
    return csms


def __read_scout_crosslinks(data: pd.DataFrame) -> List[Dict[str, Any]]:
    r"""Reads crosslinks from a Scout crosslink/residue pair result.

    Parameters
    ----------
    data : pandas.DataFrame
        The Scout crosslink/residue pair result data.

    Returns
    -------
    list of dict
        The read crosslinks.

    Notes
    -----
    This function should not be called directly, it is called from ``read_scout()``.
    """
    crosslinks = list()
    xl = data.dropna(axis=0, subset=["Alpha peptide", "Beta peptide"])
    for i, row in tqdm(
        xl.iterrows(), total=xl.shape[0], desc="Reading Scout crosslinks..."
    ):
        crosslink = create_crosslink(
            peptide_a=format_sequence(str(row["Alpha peptide"])),
            xl_position_peptide_a=int(row["Alpha peptide position"]),
            proteins_a=[
                protein.strip()
                for protein in str(row["Alpha protein mapping(s)"]).split(";")
            ],
            xl_position_proteins_a=[
                int(pos) for pos in str(row["Alpha protein(s) position(s)"]).split(";")
            ],
            decoy_a=get_bool_from_value(row["IsDecoy"]),
            peptide_b=format_sequence(str(row["Beta peptide"])),
            xl_position_peptide_b=int(row["Beta peptide position"]),
            proteins_b=[
                protein.strip()
                for protein in str(row["Beta protein mapping(s)"]).split(";")
            ],
            xl_position_proteins_b=[
                int(pos) for pos in str(row["Beta protein(s) position(s)"]).split(";")
            ],
            decoy_b=get_bool_from_value(row["IsDecoy"]),
            score=float(row["Score"]),
            additional_information={"source": __serialize_pandas_series(row)},
        )
        crosslinks.append(crosslink)
    return crosslinks


def read_scout(
    files: str | List[str] | BinaryIO,
    crosslinker: str,
    crosslinker_mass: Optional[float] = None,
    parse_modifications: bool = True,
    modifications: Dict[str, Tuple[str, float]] = SCOUT_MODIFICATION_MAPPING,
    sep: str = ",",
    decimal: str = ".",
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Read a Scout result file.

    Reads a Scout filtered or unfiltered crosslink-spectrum-matches result file or crosslink/residue pair result file in ``.csv`` format
    and returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the Scout result file(s) or a file-like object/stream.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    crosslinker_mass : float, or None, default = None
        Monoisotopic delta mass of the crosslink modification. If the crosslinker is
        defined in parameter "modifications" this can be omitted.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modifications' parameter.
    modifications : dict of str, tuple, default = ``constants.SCOUT_MODIFICATION_MAPPING``
        Mapping of Scout sequence elements (e.g. ``"+15.994900"``) and modifications (e.g ``"Oxidation of Methionine"``)
        to their modifications (e.g. ``("Oxidation", 15.994915)``).
    sep : str, default = ","
        Seperator used in the ``.csv`` file.
    decimal : str, default = "."
        Character to recognize as decimal point.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslinks or crosslink-spectrum-matches.
    KeyError
        If the specified crosslinker could not be found/mapped.
    TypeError
        If parameter verbose was not set correctly.

    Notes
    -----
    Uses ``AlphaScore`` as the score for the alpha peptide, ``BetaScore`` as the score of the
    beta peptide, and ``XLScore`` as the score of the crosslink-spectrum-match for unfiltered
    crosslink-spectrum-matches. Uses ``Score`` as the score of the crosslink-spectrum-match for
    filtered crosslink-spectrum-matches, alpha and beta peptide scores are ``None`` for filtered
    crosslink-spectrum-matches. Uses ``Score`` as the score of the crosslink for residue pairs.
    These scores should not be used for validation as Scout does it's own FDR estimation based
    on multiple scores.
    See here:
    `github.com/diogobor/Scout <https://github.com/diogobor/Scout/issues/15>`_.

    Warnings
    --------
    - When reading unfiltered crosslink-spectrum-matches, no protein crosslink positions or protein peptide positions are
      available, as these are not reported. If needed they should be annotated with ``transform.reannotate_positions()``.
    - When reading filtered crosslink-spectrum-matches, Scout does not report if the individual peptides in a crosslink are
      from the target or decoy database. The parser assumes that both peptides from a target crosslink-spectrum-match are
      from the target database, and vice versa, that both peptides are from the decoy database if it is a decoy crosslink-spectrum-match.
      This leads to only TT and DD matches, which needs to be considered for FDR estimation.
    - When reading crosslinks / residue pairs, Scout does not report if the individual peptides in a crosslink are from the
      target or decoy database. The parser assumes that both peptides from a target crosslink are from the target database,
      and vice versa, that both peptides are from the decoy database if it is a decoy crosslink. This leads to only TT and DD
      matches, which needs to be considered for FDR estimation.

    Examples
    --------
    >>> from pyXLMS.parser import read_scout
    >>> csms_unfiltered = read_scout("data/scout/Cas9_Unfiltered_CSMs.csv")

    >>> from pyXLMS.parser import read_scout
    >>> csms_filtered = read_scout("data/scout/Cas9_Filtered_CSMs.csv")

    >>> from pyXLMS.parser import read_scout
    >>> crosslinks = read_scout("data/scout/Cas9_Residue_Pairs.csv")
    """
    ## check input
    _ok = check_input(crosslinker, "crosslinker", str)
    _ok = (
        check_input(crosslinker_mass, "crosslinker_mass", float)
        if crosslinker_mass is not None
        else True
    )
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = check_input(modifications, "modifications", dict, tuple)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    _ok = check_input(verbose, "verbose", int)
    if crosslinker_mass is None:
        if crosslinker not in modifications:
            if parse_modifications:
                raise KeyError(
                    "Cannot infer crosslinker mass because crosslinker is not defined in "
                    "parameter 'modifications'. Please specify crosslinker mass manually!"
                )
            else:
                crosslinker_mass = 0.0
        else:
            crosslinker_mass = modifications[crosslinker][1]
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")

    ## data structures
    crosslinks = list()
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    for input in inputs:
        ## reading data
        data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
        ## detect input file type
        scout_file_type = detect_scout_filetype(data)
        ## process data
        if scout_file_type == "scout_csms_unfiltered":
            csms += __read_scout_csms_unfiltered(
                data,
                crosslinker,
                crosslinker_mass,
                parse_modifications,
                modifications,
                verbose,
            )
        elif scout_file_type == "scout_csms_filtered":
            csms += __read_scout_csms_filtered(
                data,
                crosslinker,
                crosslinker_mass,
                parse_modifications,
                modifications,
                verbose,
            )
        else:
            crosslinks += __read_scout_crosslinks(data)

    ## check results
    if len(crosslinks) + len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches or crosslinks were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="Scout",
        csms=csms if len(csms) > 0 else None,
        crosslinks=crosslinks if len(crosslinks) > 0 else None,
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import aggregate
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = aggregate(msannika["crosslink-spectrum-matches"])
    assert len(msannika) == 235
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = aggregate(maxquant["crosslink-spectrum-matches"])
    assert len(maxquant) == 226
    crosslinks_intersection = intersection(msannika, maxquant)
    assert len(crosslinks_intersection) == 206


def test2():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import aggregate
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = aggregate(msannika["crosslink-spectrum-matches"])
    assert len(msannika) == 235
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = aggregate(maxquant["crosslink-spectrum-matches"])
    assert len(maxquant) == 226
    crosslinks_intersection = intersection(msannika, maxquant)
    len(crosslinks_intersection) == 206
    plink = pipeline(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
        engine="pLink",
        crosslinker="DSS",
    )
    plink = aggregate(plink["crosslink-spectrum-matches"])
    assert len(plink) == 252
    crosslinks_intersection = intersection(crosslinks_intersection, plink)
    assert len(crosslinks_intersection) == 203


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import intersection

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'use' has to be one 'better_score', 'data_a', or 'data_b'! Option 'better_score' will return the intersection "
        r"with the better score \(if scores are available\)\. Option 'data_a' will return the intersection using elements of 'data_a'\. "
        r"Option 'data_b' will return the intersection using elements of 'data_b'\."
    )
    with pytest.raises(TypeError, match=err_str):
        _u = intersection(
            pr["crosslink-spectrum-matches"],
            pr["crosslink-spectrum-matches"],
            use="Set A",
        )


def test4():
    from pyXLMS.parser import read
    from pyXLMS.transform import intersection

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
        r"peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
    )
    with pytest.raises(TypeError, match=err_str):
        _u = intersection(
            pr["crosslink-spectrum-matches"],
            pr["crosslink-spectrum-matches"],
            by="sequence",
        )


def test5():
    from pyXLMS.parser import read
    from pyXLMS.transform import intersection

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
        r"-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
    )
    with pytest.raises(TypeError, match=err_str):
        _u = intersection(
            pr["crosslink-spectrum-matches"],
            pr["crosslink-spectrum-matches"],
            score="lower",
        )


def test6():
    from pyXLMS.parser import read
    from pyXLMS.transform import intersection

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = r"Verbose level has to be one of 0, 1, or 2!"
    with pytest.raises(TypeError, match=err_str):
        _u = intersection(
            pr["crosslink-spectrum-matches"],
            pr["crosslink-spectrum-matches"],
            verbose=3,
        )


def test7():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import aggregate
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = aggregate(msannika["crosslink-spectrum-matches"])
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = maxquant["crosslink-spectrum-matches"]
    err_str = r"Parameters 'data_a' and 'data_b' have to be of the same data type!"
    with pytest.raises(TypeError, match=err_str):
        _u = intersection(msannika, maxquant)


def test8():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import aggregate
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = aggregate(msannika["crosslink-spectrum-matches"])
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = aggregate(maxquant["crosslink-spectrum-matches"])
    maxquant[-5]["completeness"] = "partial"
    maxquant[-5]["score"] = None
    err_str = (
        r"Can't intersect based on score because not all data have associated scores!"
    )
    with pytest.raises(ValueError, match=err_str):
        _u = intersection(msannika, maxquant)


def test9():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = msannika["crosslink-spectrum-matches"]
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = maxquant["crosslink-spectrum-matches"]
    u = intersection(msannika, maxquant, verbose=0)
    assert u is not None


def test10():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = msannika["crosslink-spectrum-matches"]
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = maxquant["crosslink-spectrum-matches"]
    warn_str = r"Creating intersection of crosslink-spectrum-matches\. Be sure that this makes sense for your data!"
    with pytest.warns(RuntimeWarning, match=warn_str):
        _u = intersection(msannika, maxquant)


def test11():
    from pyXLMS.pipelines import pipeline
    from pyXLMS.transform import intersection

    msannika = pipeline(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    msannika = msannika["crosslink-spectrum-matches"]
    maxquant = pipeline(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    maxquant = maxquant["crosslink-spectrum-matches"]
    err_str = (
        r"Can't create intersection of crosslink-spectrum-matches for verbose level 2!"
    )
    with pytest.raises(RuntimeError, match=err_str):
        _u = intersection(msannika, maxquant, verbose=2)
